# -*- coding: utf-8 -*-
"""
/***************************************************************************
 CreateLayerFromSelectedFeatures
                                 A QGIS plugin
 This plugin creates a layer using selected features.
 Generated by Plugin Builder: http://g-sherman.github.io/Qgis-Plugin-Builder/
                              -------------------
        begin                : 2021-07-06
        git sha              : $Format:%H$
        copyright            : (C) 2021 by Murat Çalışkan
        email                : caliskan.murat.20@gmail.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""
from qgis.PyQt.QtCore import QSettings, QTranslator, QCoreApplication
from qgis.PyQt.QtGui import QIcon
from qgis.PyQt.QtWidgets import QAction

from qgis.utils import iface
from qgis.core import Qgis, QgsProject, QgsDataSourceUri

# Initialize Qt resources from file resources.py
from .resources import *
# Import the code for the dialog
from .create_layer_from_selected_features_dialog import CreateLayerFromSelectedFeaturesDialog
import os.path

import psycopg2 as ps
import sqlite3


class CreateLayerFromSelectedFeatures:
    """QGIS Plugin Implementation."""

    def __init__(self, iface):
        """Constructor.

        :param iface: An interface instance that will be passed to this class
            which provides the hook by which you can manipulate the QGIS
            application at run time.
        :type iface: QgsInterface
        """
        # Save reference to the QGIS interface
        self.iface = iface
        # initialize plugin directory
        self.plugin_dir = os.path.dirname(__file__)
        # initialize locale
        locale = QSettings().value('locale/userLocale')[0:2]
        locale_path = os.path.join(
            self.plugin_dir,
            'i18n',
            'CreateLayerFromSelectedFeatures_{}.qm'.format(locale))

        if os.path.exists(locale_path):
            self.translator = QTranslator()
            self.translator.load(locale_path)
            QCoreApplication.installTranslator(self.translator)

        # Declare instance attributes
        self.actions = []
        self.menu = self.tr(u'&Create Layer From Selected Features')

        # Check if plugin was started the first time in current QGIS session
        # Must be set in initGui() to survive plugin reloads
        self.first_start = None

    # noinspection PyMethodMayBeStatic
    def tr(self, message):
        """Get the translation for a string using Qt translation API.

        We implement this ourselves since we do not inherit QObject.

        :param message: String for translation.
        :type message: str, QString

        :returns: Translated version of message.
        :rtype: QString
        """
        # noinspection PyTypeChecker,PyArgumentList,PyCallByClass
        return QCoreApplication.translate('CreateLayerFromSelectedFeatures', message)


    def add_action(
        self,
        icon_path,
        text,
        callback,
        enabled_flag=True,
        add_to_menu=True,
        add_to_toolbar=True,
        status_tip=None,
        whats_this=None,
        parent=None):
        """Add a toolbar icon to the toolbar.

        :param icon_path: Path to the icon for this action. Can be a resource
            path (e.g. ':/plugins/foo/bar.png') or a normal file system path.
        :type icon_path: str

        :param text: Text that should be shown in menu items for this action.
        :type text: str

        :param callback: Function to be called when the action is triggered.
        :type callback: function

        :param enabled_flag: A flag indicating if the action should be enabled
            by default. Defaults to True.
        :type enabled_flag: bool

        :param add_to_menu: Flag indicating whether the action should also
            be added to the menu. Defaults to True.
        :type add_to_menu: bool

        :param add_to_toolbar: Flag indicating whether the action should also
            be added to the toolbar. Defaults to True.
        :type add_to_toolbar: bool

        :param status_tip: Optional text to show in a popup when mouse pointer
            hovers over the action.
        :type status_tip: str

        :param parent: Parent widget for the new action. Defaults None.
        :type parent: QWidget

        :param whats_this: Optional text to show in the status bar when the
            mouse pointer hovers over the action.

        :returns: The action that was created. Note that the action is also
            added to self.actions list.
        :rtype: QAction
        """

        icon = QIcon(icon_path)
        action = QAction(icon, text, parent)
        action.triggered.connect(callback)
        action.setEnabled(enabled_flag)

        if status_tip is not None:
            action.setStatusTip(status_tip)

        if whats_this is not None:
            action.setWhatsThis(whats_this)

        if add_to_toolbar:
            # Adds plugin icon to Plugins toolbar
            self.iface.addToolBarIcon(action)

        if add_to_menu:
            self.iface.addPluginToVectorMenu(
                self.menu,
                action)

        self.actions.append(action)

        return action

    def initGui(self):
        """Create the menu entries and toolbar icons inside the QGIS GUI."""

        icon_path = ':/plugins/create_layer_from_selected_features/icon.png'
        self.add_action(
            icon_path,
            text=self.tr(u''),
            callback=self.run,
            parent=self.iface.mainWindow())

        # will be set False in run()
        self.first_start = True


    def unload(self):
        """Removes the plugin menu item and icon from QGIS GUI."""
        for action in self.actions:
            self.iface.removePluginVectorMenu(
                self.tr(u'&Create Layer From Selected Features'),
                action)
            self.iface.removeToolBarIcon(action)


    def run(self):
        self.layers = [i for i in iface.layerTreeView().selectedLayers() if i.type() == 0]
        if self.layers:        
            self.selectedFeatures = {lyr.name():list(lyr.getSelectedFeatures()) for lyr in self.layers}
            self.selectedFeatureIds = {name:[f.id() for f in feat] for (name,feat) in self.selectedFeatures.items()}            
            
            for layer in self.layers:
                self.selected_feat_ids = self.selectedFeatureIds[layer.name()]
                
                if self.selected_feat_ids:
                    self.lyr_new = layer.clone()
                    self.mapLayers = [i.name() for i in QgsProject.instance().mapLayers().values()]
                    self.base_name = self.new_name = layer.name() + "_filtered"
                    self.suff=1
                    
                    while self.new_name in self.mapLayers:
                        self.new_name = self.base_name + "_" + str(self.suff)
                        self.suff +=1
                    self.lyr_new.setName(self.new_name)
                    QgsProject.instance().addMapLayer(self.lyr_new)
                    
                    if layer.storageType() in ('Memory storage', 'Delimited text file'):
                        if len(self.selected_feat_ids) > 1:
                            self.lyr_new.setSubsetString('$id IN {}'.format(tuple(self.selected_feat_ids)))
                        else:
                            self.lyr_new.setSubsetString('$id = {}'.format(self.selected_feat_ids[0]))
                            
                    elif layer.storageType() in ('PostgreSQL database with PostGIS extension',):
                        self.connInfo = QgsDataSourceUri(layer.dataProvider().dataSourceUri())
                        self.dbname = self.connInfo.database()
                        self.host = self.connInfo.host()
                        self.port = self.connInfo.port()
                        self.user = self.connInfo.username()
                        self.password = self.connInfo.password()    
                        self.pk_field = self.connInfo.keyColumn()
                        
                        if self.pk_field == "tid":
                            self.iface.messageBar().pushMessage("Warning", f"The layer '{layer.name()}' doesn't have primary key field!" , level=Qgis.Warning, duration=5)
                            QgsProject.instance().removeMapLayer(self.lyr_new.id())
                        
                        else:
                            self.selectedFeaturePKs = [f[self.pk_field] for f in self.selectedFeatures[layer.name()]]
                            
                            if len(self.selected_feat_ids) > 1:
                                self.lyr_new.setSubsetString(self.pk_field + ' IN {}'.format(tuple(self.selectedFeaturePKs)))
                            else:
                                self.lyr_new.setSubsetString(self.pk_field + ' = {}'.format(self.selectedFeaturePKs[0]))
                    
                    elif layer.storageType() in ('SQLite',):
                        self.path, self.layer_name = layer.source().split("|")
                        self.layer_name = self.layer_name.split("=")[-1]
                        self.query = f"PRAGMA table_info('{self.layer_name}')"
                        self.conn = sqlite3.connect(self.path)                        
                        self.cur = self.conn.cursor()
                        self.cur.execute(self.query)                        
                        self.pk_field = [i[1] for i in self.cur.fetchall() if i[-1] == 1][0]
                        
                        self.selectedFeaturePKs = [f[self.pk_field] for f in self.selectedFeatures[layer.name()]]
                        
                        if len(self.selected_feat_ids) > 1:
                            self.lyr_new.setSubsetString(self.pk_field + ' IN {}'.format(tuple(self.selectedFeaturePKs)))
                        else:
                            self.lyr_new.setSubsetString(self.pk_field + ' = {}'.format(self.selectedFeaturePKs[0]))   
                            
                    else:
                        if len(self.selected_feat_ids) > 1:
                            self.lyr_new.setSubsetString('"fid" IN {}'.format(tuple(self.selected_feat_ids)))
                        else:
                            self.lyr_new.setSubsetString('fid = {}'.format(self.selected_feat_ids[0]))
